;			The Binder Application
;
; It is responsible for compiling Binder assertions and maintaining the
; application namespace and other namespaces.
;
; binder-application-init INIT-FNAME -> NS-CACHE
; Initialize the binder application and load the initial assertion from the
; INIT-FNAME into the system namespace. Return the NS-CACHE
; that contains compiled assertions for the application and system
; namespaces.
;
; binder-application-query REQ NS-CACHE -> BOOL
; Given a request, decide if the request should be allowed.
; REQ is an assoc list, of (key STRING-value) pairs.
; The following keys are expected:
;       Channel - the name of the channel
;       Access  - "read" or "write"
;	IP - the IP address of the client (dot-quad-notation, as a string)
;            becomes the answer of "application says ipaddress" predicate
;       Subject - the pubkey or other identifier of the subject.
;            It becomes the answer of "application says pubkey_fingerprint"
;            predicate
;
; binder-application-extend NS-CACHE NAME ASSERTION -> NS-CACHE-NEW or #f
; Compile the ASSERTION and add it to NS-CACHE under the key NAME.
; Return #f if the compilation fails.
;
; NS-CACHE is an abstract collection with the regular methods
; empty, get and extend. 
; The key is the namespace name converted to a safe symbol
; (see binder-serialize)
;
; All errors go to stderr.
;
; $Id: binder-application.scm,v 1.2 2004/10/14 22:38:11 oleg Exp $

(module binder-application
  (include "myenv-bigloo.scm")
  (include "srfi-13-local.scm")
  (include "util.scm")
  (include "char-encoding.scm")
  (include "input-parse.scm")
  (include "binder-kanren.scm")
  (import
    (binder-compile "binder-compile.scm")
  )
  (export
    (binder-application-init::struct init-fname::bstring)
    (binder-application-query::bbool req::pair-nil ns-cache::struct)
    (binder-application-extend ns-cache::struct name::bstring 
      assertion::bstring)
    )
)

; The namespace cache. Currently it is an assoc list

(define-record-type namespace-cache
  (make-namespace-cache val)
  namespace-cache?
  (val namespace-cache-val))


(define (empty-namespace-cache)
  (make-namespace-cache '()))

(define (namespace-cache-get cache ns-name)
  (assert (symbol? ns-name) (namespace-cache? cache))
  (cond
    ((assq ns-name
       (namespace-cache-val cache)) => cdr)
    (else #f)))

; Extend the cache. Replace the existing association, if any
(define (namespace-cache-put cache ns-name assertion)
  (assert (symbol? ns-name) (namespace-cache? cache))
  (let ((cache-val (namespace-cache-val cache)))
    (make-namespace-cache
      (cons (cons ns-name assertion)
	(cond
	  ((assq ns-name cache-val) =>
	    (lambda (m)
	      (let loop ((cache-val cache-val))
		(if (eq? m (car cache-val)) (cdr cache-val)
		  (cons (car cache-val) (loop (cdr cache-val)))))))
	  (else cache-val))))))

	

; Distinguished namespace names

(define ns-name-application (binder-serialize "application"))
(define ns-name-system (binder-serialize "system"))

;------------------------------------------------------------------------
;			The application Namespace

; Primitive operations of the application namespace
(define application-primitives
  (extend-relation (x)
    ; ip_of/2
    (relation (ip ipnet)
      (to-show `(ip_of/2 ,ip ,ipnet))
      (all!
	(project/no-check (ip ipnet)
	  (predicate 
	    (and
	      (IP-address? ip)
	      (IP-address? ipnet)
	      ; match min-width higher-order bits of ip and ipnet
	      (let ((min-width (min (ip-addr-nbits ip)
				 (ip-addr-nbits ipnet))))
		(case min-width
		  ((0) #t)		; no bits
		  ((32) (equal? (ip-addr-ip ip) (ip-addr-ip ipnet))) ; all bits
		  (else
		    (let
		      ((shift-f (expt 2 (- 32 min-width))))
		      (= (floor (/ (ip-addr-ip ip) shift-f))
			 (floor (/ (ip-addr-ip ipnet) shift-f))))))))))))
    ))
	      


;------------------------------------------------------------------------
;			The Decision engine

; Take an assertion and return the compiled assertion or #f on failure
; The compiled assertion needs to be applied to the ns-entailment
; function to yeild the entailment relation for the namespace

(define (compile-assertion str)
  (cerr nl "Compiling the assertion " str "..." nl)
  (let ((result (binder-compile str)))
    (cond 
      ((error-assertion? result)
	(cerr nl "Error compiling the assertion: "
	  (error-assertion-msg result))
	#f)
      ((null? result)
	(cerr nl "Empty assertion" nl)
	#f)
      (else
	;(pp result)
	(eval result (interaction-environment))))))


(define (binder-application-init init-fname)
  (cerr nl "Initializing Binder..." nl
    "Loading the initial assertion from " init-fname nl)
  (let ((compiled-assertion
	  (compile-assertion
	    (call-with-input-file init-fname
	      (lambda (port)
		(next-token '() '(*eof*) "reading string" port))))))
    (if compiled-assertion
      (namespace-cache-put (empty-namespace-cache) 
	ns-name-system compiled-assertion)
      (begin
	(cerr nl "Failed to load the initial assertion. Exiting" nl)
	(exit 4)))))

(define init-ns-cache
  (binder-application-init "binder-init-assertion.txt"))


; binder-application-extend NS-CACHE NAME ASSERTION -> NS-CACHE-NEW or #f
(define (binder-application-extend cache name assertion)
  (cerr nl "Binder extend: " name " ... ")
  (and-let*
    ((compiled-assertion
       (compile-assertion assertion))
      (key (binder-serialize name))
      ((not (memq key `(,ns-name-system ,ns-name-application))))
      )
    (namespace-cache-put cache key compiled-assertion)
    ))

; The main query application
(define (binder-application-query cache req)
  (let
    ((channel (binder-serialize (lookup-def 'Channel req)))
     (access  (binder-serialize (lookup-def 'Access req)))
     (ip      (assert (build-ip-address (lookup-def 'IP req))))
     (subject 
       (let ((name (lookup-def 'Subject req '#f)))
	 (and name (binder-serialize name))))
     (system-ns (assert (namespace-cache-get cache ns-name-system)))
      )

    ; build the application ns
    (let*
      ((application
	 (extend-relation (x)
	   (fact () `(ipaddress/1 ,ip))
	   application-primitives))
	(application
	  (if subject
	    (extend-relation (x)
	      (fact () `(pubkey_fingerprint/1 ,subject))
	      application)
	    application))

	(application
	  (extend-relation (x)
	    (relation (head-let x)
	      (all!
		(trace-vars 'application-ns-query (x))
		fail))
	    application))
	)

      (letrec
	((application-ns
	   (lambda (ns-name)
	     (cerr nl "querying app ns for " ns-name nl)
	     (cond 
	       ((eq? ns-name ns-name-application)
		 application)
	       ((namespace-cache-get cache ns-name) =>
		 (lambda (assertion)
		   (assertion application-ns)))
	       (else
		 (cerr "not found the named assertion" nl)
		 (lambda (x) fail))))))

	(not (not
	       (solution () ((system-ns application-ns)
			      `(may/3 channel ,channel ,access)))))
      ))))


; Scenario Local-MEMO
(assert
  (binder-application-query init-ns-cache
   '((Channel "MEMO")
     (Access write)
     (IP "10.10.1.1")
     )))

(assert
  (binder-application-query init-ns-cache
   '((Channel "MEMO")
     (Access read)
     (IP "192.168.5.4")
     )))

; Scenario Remote-MEMO
(assert
  (not
    (binder-application-query init-ns-cache
      '((Channel "MEMO")
	 (Access read)
	 (IP "172.10.10.1")
	 ))))

(assert
  (not
   (binder-application-query init-ns-cache
   '((Channel "MEMO")
     (Access read)
     (Subject "abcdef") ; Dean
     (IP "172.10.10.1")
     ))))

(assert
  (binder-application-query init-ns-cache
   '((Channel "MEMO")
     (Access write)
     (Subject "0123456789") ;Joe
     (IP "172.10.10.1")
     )))

; Scenario DEMO-IMG-simple
(assert
 (not
  (binder-application-query init-ns-cache
   '((Channel "DEMO-IMG")
     (Access read)
     (IP "10.0.0.1")
     ))))

(assert
 (not
  (binder-application-query init-ns-cache
   '((Channel "DEMO-IMG")
     (Access read)
     (IP "10.0.0.1")
     (Subject "abcdef") ; Dean
     ))))

(let ((cache
	(binder-application-extend init-ns-cache
	  "abcdef" ;Dean
	  "may(?,?,?).")))
  (assert (binder-application-query cache
   '((Channel "DEMO-IMG")
     (Access read)
     (IP "172.10.10.1")
     (Subject "abcdef") ; Dean
     ))))

(let ((cache
	(binder-application-extend init-ns-cache
	  "abcdef" ;Dean
	  (string-append
	    "may(channel,?,read) :- known_user(Eric)." nl
	    "known_user(Eric) :- 
               application says pubkey_fingerprint(\"dddddd\")." nl))))
  (assert
    (not 
      (binder-application-query cache
	'((Channel "DEMO-IMG")
	   (Access write)
	   (IP "172.10.10.1")
	   (Subject "dddddd") ; Eric
	   )))))

