;		Compiler of Binder assertions into KANREN
;
; The syntax is described in
; http://www.metnet.navy.mil/Metcast/Auth-use-cases.html
;
;	binder-compile STR -> SCHEME-CODE-EXPR or ERROR-ASSERTION
;
; The function binder-compile takes an assertion, as string, and
; returns either an S-expression for the compiled Scheme (KANREN)
; code, or an ERROR assertion.
; The S-expression is ready to be passed to EVAL. The result of
; evaluating the S-expression is an essentially closed procedure
; that takes one argument:
;   _ns-entails : a curried procedure of two arguments, that takes
;   the namespace name (a string) and an atom in KANREN representation.
;   The result is the KANREN antecedent (which fails or succeeds
;   depending on if the atom succeeds in the specified namespace).
;
; When the result of evaluating (binder-compile STR) is applied to
; _ns-entails, the result is the compiled assertion proper: a
; procedure of one argument.  The argument is the atom to prove. As
; customery in KANREN, an atom may(x,y,?) is represented as
; `(may/3 x y ,_)
;
; For example, the following assertion:

; may(channel,MEMO,?) :- application says ipaddress(?IP),
;                        internal(?IP).
; may(channel,DEMO-IMG, ?r) :-
;	Administrator says pubkey(Dean,?Dean_key),
;	?Dean_key says may(channel,DEMO-IMG, ?r),
;	may(channel,MEMO,Read).
; internal(#p192.16.167.26).
;
; is compiled into
;
; (lambda (_ns-entails)
;   (let ((_adm456 (_ns-entails 'Administrator))
;         (_app457 (_ns-entails 'application)))
;     (letrec ((_g12345
;        (extend-relation-with-recur-limit 100 (x)
; 	 (relation ()
; 	   (to-show `(may/3 channel MEMO ,_))
; 	   (exists (IP)
; 	     (all (_app457 `(ipaddress/1 ,IP))
; 	       (_g12345 `(internal/1 ,IP)))))
; 	 (relation (r)
; 	   (to-show `(may/3 channel DEMO-IMG ,r))
; 	   (exists (Dean_key)
; 	     (all
; 	       (_adm456 `(pubkey/2 Dean ,Dean_key))
; 	       (project (Dean_key)
; 	       (all
; 	         ((_ns-entails Dean_key) `(may/3 channel DEMO-IMG ,r))
; 	         (_g12345 `(may/3 channel MEMO Read))))))
; 	 (fact () `(internal/1 "192.16.167.26/32")))))
;       _g12345)))
;
; All constants and variable names from assertion are to be passed
; through the procedure binder-serialize (below). Currently the procedure
; represents string constants as symbolic constants. The procedure
; encodes `bad' characters using the conventional URL encoding (as %xx).
; Note that characters #\% and #\_ are considered `bad'.
;
; The 'project' clause was emitted as the result of a "strictness analysis"
; (because _ns-entails takes a ground term -- a string or a symbol.
; So, if _ns-entails is applied to a variable, that variable must first
; be resolved).
;
; Note for the future:
; Conjunctions in the body of the rule are appropriately
; reshuffled: we make sure that every atom whose context is a variable
; comes after an atom whose FV list includes that variable. 
;	    
; $Id: binder-compile.scm,v 1.1 2004/07/10 04:15:56 oleg Exp $

(module binder-compile
  (include "myenv-bigloo.scm")
  (include "srfi-13-local.scm")
  (include "util.scm")
  (include "input-parse.scm")
  (import
    (binder-parse "binder-parse.scm")
  )
  (export
    (binder-compile assertion::bstring)
    (error-assertion?::bbool err::struct)
    (error-assertion-msg::pair err::struct)
    (binder-serialize obj)
    )
)

; The limit on the recursion depth in any namespace
(define binder-recursion-limit 100)


; Left fold combinator for a single list
(define (fold kons knil lis1)
  (let lp ((lis lis1) (ans knil))
    (if (null? lis) ans
      (lp (cdr lis) (kons (car lis) ans)))))

; Given an assoc list, insert a new assoc of a KEY to VALUE
; (replacing the old one, if existed)
(define (insert-assoc alist key val)
  (if (assoc key alist)
    (let loop ((lst alist))
      (if (equal? key (caar lst))
	(cons (cons key val) (cdr lst))
	(cons (car lst) (loop (cdr lst)))))
    (cons (cons key val) alist)))

; Serialize an object into the external representation suitable
; to be used in KANREN code. See the comments above -- especially
; about converting string constants into symbolic constants while
; safely quoting bad characters
; The anonymous variable is serialized as _

(define (binder-serialize obj)
  ; A safe symbol includes only alphanumeric characters and safe
  ; special characters. Everything unsafe is URL (%xx) quoted.
  ; Characters #\% and #\_ are not safe.
  (define (string->safe-symbol str)
    (assert (positive? (string-length str)))
    (let ((safe-special "!@~^&*></-+=."))
      (string->symbol
	(call-with-output-string 
	  (lambda (oport)
	    (call-with-input-string str
	      (lambda (iport)
		(do ((c (read-char iport) (read-char iport)))
		    ((eof-object? c))
		  (cond
		    ((char-alphabetic? c) (write-char c oport))
		    ((char-numeric? c) (write-char c oport))
		    ((string-index safe-special c) (write-char c oport))
		    (else
		      (write-char #\% oport)
		      (display (number->string (char->integer c) 16)
			oport)))))))))))
  (cond
    ((symbol? obj) obj)
    ((string? obj) (string->safe-symbol obj))
    ((binder-var? obj)
      (let ((name (binder-var-name obj)))
	(if (string=? name "") '_
	  (string->safe-symbol name))))
    ((IP-address? obj) obj)
    (else
      (assert #f report: "Bad obj for binder-serialize: " obj))))

(let ((result (binder-serialize "_"))) (assert (equal? result '%5f)))
(let ((result (binder-serialize "abC"))) (assert (equal? result 'abC)))
(let ((result (binder-serialize "%abC"))) (assert (equal? result '%25abC)))
(let ((result (binder-serialize 'may/3))) (assert (equal? result 'may/3)))

; Generate a unique and distinct symbol
; To make sure the symbol would never clash with the symbols
; in the binder assertions, we prepend _ to the symbol name.
; Symbols created by binder-serialize can never start with an underscore.
(define (binder-gensym)
  (symbol-append '_ (gensym)))


; collect-constant-ns-names (RULE-NEST ...) -> ((NS-NAME . NS-SYMBOL)...)
; Scan the list of rule-nests searching for the names of constant
; namespaces (NS-NAME, a string), generate unique symbolic
; names for them, and return the corresponding translation dictionary

(define (collect-constant-ns-names nests)
  (let ((ns-names
	  (fold
	    (lambda (nest acc)
	      (fold
		(lambda (clause acc)
		  (fold
		    (lambda (pcall acc)
		      (let ((ctx (assertion-pcall-ctx pcall)))
			(if (and (string? ctx) ; a constant name
			      (not (member ctx acc))) ; did not occur before
			  (cons ctx acc)
			  acc)))
		    acc
		    (assertion-clause-body clause)))
		acc
		(cdr nest)))
	    '()
	    nests)))
    (map (lambda (ns-name) (cons ns-name (binder-gensym))) ns-names)))


; Compile a clause: clause SELF NS-ENTAILS NS-NAMES-DICT
; SELF is the name of the identifier bound to the current assertion.
; We use SELF for compiling recursive calls, that is, for atoms
; whose context is #f
; NS-ENTAILS is the name of the identifier that will have been
; bound to the entailment function
; (which checks the entailment of atoms in other namespaces). We use
; NS-ENTAILS when compiling atoms whose context is variable.
; NS-NAMES-DICT is the translation dictionary for constant namespaces

(define (compile-clause clause self ns-entails ns-names-dict)
  ; collect logical vars that occur in the body but not in the head
  (define (collect-existential-vars clause)
    (let ((head-vars (assertion-clause-fv clause)))
      (fold
	(lambda (pcall acc)
	  (fold
	    (lambda (var acc)
	      (if (and (not (member var acc))
		    (not (member var head-vars)))
		(cons var acc)
		acc))
	    acc
	    (assertion-pcall-fv pcall))) ; anon variables are never mentioned
	'()
	(assertion-clause-body clause))))
  (define (compile-atom pred-name args)
    (list 'quasiquote
      (cons pred-name
	(map (lambda (term)
	       (if (binder-var? term)
		 `(,'unquote ,(binder-serialize term))
		 (binder-serialize term)))
	  args))))
  (define (compile-pcall pcall)
    (let ((ctx (assertion-pcall-ctx pcall))
	  (compiled-atom (compile-atom
			   (assertion-pcall-name pcall)
			   (assertion-pcall-args pcall))))
      (cond
	((not ctx)			; local pcall
	  `(,self ,compiled-atom))
	((binder-var? ctx)
	  `((,ns-entails ,(binder-serialize ctx)) ,compiled-atom))
	((assoc ctx ns-names-dict) =>
	  (lambda (ass)
	    `(,(cdr ass) ,compiled-atom)))
	(else (assert #f 'compile-pcall)))))
  ; Introduces project -- the strictness annotation
  ; list of pcall -> list of compiled pcalls
  (define (analyze-compile-pcalls projected-vars pcalls)
    (if (null? pcalls) '()
      (let ((ctx (assertion-pcall-ctx (car pcalls))))
	(if (and (binder-var? ctx)
	      (not (member ctx projected-vars)))
	  `((project (,(binder-serialize ctx))
	      (all
		,(compile-pcall (car pcalls))
		,@(analyze-compile-pcalls (cons ctx projected-vars)
		    (cdr pcalls)))))
	  (cons (compile-pcall (car pcalls))
	    (analyze-compile-pcalls projected-vars (cdr pcalls)))))))
  (define (compile-fact clause)
    `(fact
       ,(map binder-serialize (assertion-clause-fv clause))
       ,(compile-atom (assertion-clause-name clause)
	  (assertion-clause-head clause))))
  (define (compile-rel clause)
    `(relation
       ,(map binder-serialize (assertion-clause-fv clause))
       (to-show ,(compile-atom (assertion-clause-name clause)
		   (assertion-clause-head clause)))
       (exists ,(map binder-serialize (collect-existential-vars clause))
	 (all ,@(analyze-compile-pcalls '() (assertion-clause-body clause))))
       ))
  (assert (not (assertion-clause-ctx clause)))
  (if (null? (assertion-clause-body clause))
    (compile-fact clause)
    (compile-rel clause))
)


;	binder-compile STR -> SCHEME-CODE-EXPR or ERROR-ASSERTION
(define (binder-compile str)
  (let ((parsed (binder-parse str)))
    (if (or (error-assertion? parsed) (null? parsed)) parsed
      (let ((ns-entails (binder-gensym))
	    (self (binder-gensym))
	    (ns-names-dict (collect-constant-ns-names parsed)))
	`(lambda (,ns-entails)
	   (let ,(map			; generate let bindings for constant ns
		    (lambda (ns-name-symb)
		      `(,(cdr ns-name-symb) 
			 (,ns-entails 
			   ',(binder-serialize (car ns-name-symb)))))
		    ns-names-dict)
	     (letrec ((,self
			(extend-relation-with-recur-limit
			  ,binder-recursion-limit (x)
			  ,@(map 
			      (lambda (clause)
				(compile-clause clause self ns-entails
				  ns-names-dict))
			      ; turn a list of rule-nests to list of clauses
			      (apply append
				(map (lambda (nest) (cdr nest)) parsed))))))
	       ,self)))
))))

(define binder-test-assertion-1
  (apply string-append
    (list-intersperse
      (list
	"may(channel,MEMO,?) :- application says ipaddress(?IP),"
	"                       internal(?IP)."
	"may(channel,DEMO-IMG, ?r) :-"
	"   Administrator says pubkey(Dean,?Dean_key),"
	"   ?Dean_key says may(channel,DEMO-IMG, ?r),"
	"   may(channel,MEMO,Read)."
	"internal(#p192.16.167.26).")
      nl)))



(let ((result (collect-constant-ns-names
		(binder-parse binder-test-assertion-1))))
  (assert (assoc "Administrator" result)
          (assoc "application" result)))

(pp (binder-compile binder-test-assertion-1))

(define (test-binder-compile str expected)
  (cerr "Testing compilation of " str "..." nl)
  (let* ((result (binder-compile str)))
    (if (not (test-binder-match? result expected))
      (assert (equal? result expected)))))

(test-binder-compile binder-test-assertion-1
  `(lambda (*any*)
     (let ((*any* (*any* ',(binder-serialize "Administrator")))
	   (*any* (*any* ',(binder-serialize "application"))))
       (letrec ((*any*
	(extend-relation-with-recur-limit 100 (x)
	 (relation ()
	   (to-show `(may/3 channel MEMO ,_))
	   (exists (IP)
	     (all (*any* `(ipaddress/1 ,IP))
	       (*any* `(internal/1 ,IP)))))
	 (relation (r)
	   (to-show `(may/3 channel DEMO-IMG ,r))
	   (exists (Dean%5fkey)
	     (all
	       (*any* `(pubkey/2 Dean ,Dean%5fkey))
	       (project (Dean%5fkey)
		 (all
		   ((*any* Dean%5fkey) `(may/3 channel DEMO-IMG ,r))
		   (*any* `(may/3 channel MEMO Read)))))))
	 (fact () `(internal/1 *any*)))))
      *any*))))

(test-binder-compile binder-test-init-assertion
  '(lambda (*any*)
     (let ((*any* (*any* 'application)))
       (letrec ((*any*
	(extend-relation-with-recur-limit 100 (x)
	 (relation ()
	   (to-show `(may/3 channel MEMO ,_))
	   (exists (IP)
	     (all (*any* `(ipaddress/1 ,IP))
	       (*any* `(internal/1 ,IP)))))
	 (relation ()
	   (to-show `(may/3 channel MEMO ,_))
	   (exists ()
	     (all (*any* `(known_user/1 Joe)))))
	  (relation (Access)
	   (to-show `(may/3 channel DEMO-IMG ,Access))
	   (exists (Dean%5fkey)
	     (all (*any* `(pubkey/2 Dean ,Dean%5fkey))
	       (project (Dean%5fkey)
		 (all
		   ((*any* Dean%5fkey) 
		     `(may/3 channel DEMO-IMG ,Access)))))))
	  (fact () `(internal/1 *any*))
	  (relation (IP)
	    (to-show `(internal/1 ,IP))
	    (exists ()
	      (all (*any* `(ip_of/2 ,IP *any*)))))
	  (relation (user)
	    (to-show `(known_user/1 ,user))
	    (exists (key)
	      (all
		(*any* `(pubkey/2 ,user ,key))
		(*any* `(pubkey_fingerprint/1 ,key)))))
	  (fact () `(pubkey/2 Dean *any*))
	  (fact () `(pubkey/2 Joe *any*))
	  (relation (x)
	    (to-show `(pubkey_fingerprint/1 ,x))
	    (exists ()
	      (all (*any* `(pubkey_fingerprint/1 ,x)))))
	  )))
	 *any*))))

