;		Parser of Binder assertions
;
; The syntax is described in
; http://www.metnet.navy.mil/Metcast/Auth-use-cases.html
;
; The main exported function is binder-parse
;	binder-parse STRING -> (<rule-nest> ...) or <error-assertion>
;
; The function binder-parse receives an assertion as a string.  Most
; of the time assertions come in signed messages, so we have to read
; the whole assertion into a string to verify the signature. So, we'd
; rather parse from a string rather from a port. Furthermore, we don't
; want to get blocked if the port is blocking.
;
; The result is a list of <rule-nest> or <error-assertion>.
; An <error-assertion> signifies a parsing error. See the structure
; error-assertion below for more details.
;
; All parsed assertions are grouped by predicates-arity.
; <rule-nest> = (<predicate> <pred-clause> ...)
; <predicate> is a symbol whose spelling is "predicate-given-name/arity"
; <pred-clause> is a clause that corresponds to the given predicate
; (implicitly, the arity is the part of a predicate name, as it is in Prolog).
; <pred-clause> is a record, see below. It contains the list of the
; variables in the head, the head of the clause, and the list of
; conjunctions that constitute the body of the clause.
; Variables are special records. The anonymous variable is
; not included in the list of variables.
;
; We verify that all the clauses are safe: that is, a variable that
; occurs in the head of the clause must also occur in the body of the
; clause.  (anonymous variables don't count?).  Predicates at the head
; of the clause must not be qualified by a context.  
; At present, we also require that if a context of some atom is a variable,
; that variable should occur either in the head of a clause, or
; in a previous atom in a conjunction. Of course we can re-order
; atoms in a conjunction to try to satisfy the previous condition. However,
; we chose to place the burden of properly ordering conjunction onto the
; user, to avoid surprises. The condition essentially says that
; the context of an atom is in the IN mode. The context cannot be an
; anonymous variable either.

; We use Bigloo's built-in lexers and parsers. So the code is Bigloo-specific.
;
; $Id: binder-parse.scm,v 1.1 2004/07/10 04:15:56 oleg Exp $

(module binder-parse
  (include "myenv-bigloo.scm")
  (include "srfi-13-local.scm")
  (include "util.scm")
  (include "input-parse.scm")
  (export
    (binder-parse::pair-nil str::bstring)

    (build-ip-address str::bstring) ; parse the IP address
    (IP-address?::bbool ip-struct::struct)
    (ip-addr-ip ip-struct::struct)
    (ip-addr-nbits::bint ip-struct::struct)

    (binder-var?::bbool bv::struct)
    (binder-var-name::bstring bv::struct)
    binder-anon-var::struct

    (assertion-clause?::bbool clause::struct)
    (assertion-clause-name::symbol clause::struct)
    (assertion-clause-ctx clause::struct)
    (assertion-clause-fv::pair-nil clause::struct)
    (assertion-clause-head::pair clause::struct)
    (assertion-clause-body::pair-nil clause::struct)

    (assertion-pcall?::bbool pcall::struct)
    (assertion-pcall-name::symbol pcall::struct)
    (assertion-pcall-ctx pcall::struct)
    (assertion-pcall-fv::pair-nil pcall::struct)
    (assertion-pcall-args::pair pcall::struct)

    (error-assertion?::bbool err::struct)
    (error-assertion-msg::pair err::struct)

    binder-test-init-assertion::bstring

    (test-binder-match?::bbool e1::obj e2::obj)
  )
)

; From SRFI-1
(define (fold-right kons knil lis1)
    (let recur ((lis lis1))
       (if (null? lis) knil
	    (let ((head (car lis)))
	      (kons head (recur (cdr lis)))))))

; Left fold combinator for a single list
(define (fold kons knil lis1)
  (let lp ((lis lis1) (ans knil))
    (if (null? lis) ans
      (lp (cdr lis) (kons (car lis) ans)))))

;------------------------------------------------------------------------
;			Data types
; The field 'context' below describes the context of an atom.
; The context is #f (meaning the current context) or a string,
; or a variable.

; Use equal? to compare variables, including the anon variable!
(define-record-type binder-var
  (make-binder-var name)
  binder-var?
  (name binder-var-name))

(define binder-anon-var (make-binder-var "")) ; The anonymous variable

; IP address -- for host and network addresses
; The address has the form <ip-addr>/<nbits-in-network-mask>
; At present, ip-addr is an integer (long integer, in Bigloo terms)
; denoting an IPv4 address
(define-record-type IP-address
  (make-ip-address ip nbits)
  IP-address?
  (ip ip-addr-ip)
  (nbits ip-addr-nbits))

; A clause of an assertion
(define-record-type assertion-clause
  (make-assertion-clause
    predict				; the name of the head predicate
					; (including the arity)
    context				; The context name, or #f
    fv					; free variables in the predicate
					; (anon variables are not listed)
    head				; List of the variables/constants
					; that constitute the head
    body				; A list of assertion-pcall
    )
  assertion-clause?
  (predict  assertion-clause-name)
  (context    assertion-clause-ctx)
  (fv         assertion-clause-fv)
  (head       assertion-clause-head)
  (body       assertion-clause-body))

; A predicate invocation
(define-record-type assertion-pcall
  (make-assertion-pcall
    predict				; the name of the predicate
					; (including the arity)
    context				; The context name, see above
    fv					; free variables in the predicate
					; (anon variables are not listed)
    args				; List of the variables/constants
					; that constitute the pred invocation
    )
  assertion-pcall?
  (predict assertion-pcall-name)
  (context   assertion-pcall-ctx)
  (fv        assertion-pcall-fv)
  (args      assertion-pcall-args))
  



; A record that denotes a parsing error or other error occurred
; during the compilation of the assertion.
; The message provides more details about the nature
; of the error and perhaps its location
(define-record-type error-assertion
  (make-error-assertion msg)
  error-assertion?
  (msg error-assertion-msg))


;------------------------------------------------------------------------
;				Parsing

; LALR(1) Grammar of Binder
; The semantic value of the grammar (i.e., of its top predicate,
; 'assertion', is the list clauses or 'error-assertion')
(define binder-grammar
  (lalr-grammar 
    (lparen rparen dot comma conseq says variable symb-constant constant error)
    (assertion
      (() '())				; empty program
      ((statement assertion) (cons statement assertion))
      ((error) (make-error-assertion error))
      )
    (statement
      ((rule) rule)
      ((stmt-fact) stmt-fact)
      )
    (rule
      ((head conseq body dot) (gr-atom->clause head body)))
    (stmt-fact
      ((head dot) (gr-atom->clause head '())))
    (head
      ((patom) patom))			; head cannot contain any context
    (body
      ((atom) (list (gr-atom->pcall atom)))
      ((atom comma body) (cons (gr-atom->pcall atom) body)))
    (atom
      ((catom) catom)
      ((patom) patom))
    (catom
      ((context says patom) (gr-atom-context-set patom context)))
    (patom
      ((predict lparen args rparen) (gr-make-atom predict args)))
    (predict
      ((symb-constant) symb-constant))
    (args
      ((term) (list term))
      ((term comma args) (cons term args)))
    (context
      ((term) term)) ; can be a variable!
    (term
      ((symb-constant) symb-constant)
      ((constant)      constant)
      ((variable)      (make-binder-var variable)))
    ))


; Auxiliary semantic functions and data structures used in parsing

; atom: a list [context predicate . args]
; context can be #f or a constant or gr-var
(define (gr-make-atom predict args)
  (cons* #f predict args))
(define (gr-atom-context-set atom new-context)
  (cons new-context (cdr atom)))

; Convert an atom to a pcall structure
; We know that the args list has at least one element
(define (gr-atom->pcall atom)
  (let*-values
    (((ctx pred-name . args) (apply values atom))
     ((pred-name)
       (string->symbol
	 (string-append pred-name "/" (number->string (length args)))))
     ((fv)				; collect fv. ctx may be a var too
       (fold				; don't count anon vars
	 (lambda (el acc)
	   (cond
	     ((not (binder-var? el)) acc)
	     ((equal? el binder-anon-var) acc)
	     (else (cons el acc))))
	 '()
	 (cons ctx args)))
      )
    (make-assertion-pcall pred-name ctx fv args)))


; Convert an atom from the head to a clause
; body is the list of pcall structures
(define (gr-atom->clause head body)
  (let*
    ((head-str (gr-atom->pcall head)))
    (make-assertion-clause
      (assertion-pcall-name head-str)
      (assertion-pcall-ctx head-str)
      (assertion-pcall-fv head-str)
      (assertion-pcall-args head-str)
      body)))


; Lexical structure of Binder
; This regular grammar is supposed to be used with the lalr-grammar
; Therefore, the result from the grammar must be either:
;  - a symbol (naming non-terminal)
;  - a pair (SYMBOL . VALUE) where VALUE is the semantic value of
;    the non-terminal
;  - EOF
; A special symbol is error, signifying the lexing error.
; See bigloo2.4b/runtime/Read/reader.scm for examples

(define old-hygien *hygien?*)
(set! *hygien?* #f)			; regular-grammar cannot work
					; when R5RS macros are enabled...

(define binder-lex
  (regular-grammar
    (
      (letter   (in ("azAZ")))
      (special  (in "!@~$%^&*></-_+=.:"))
      (id       (: (or letter special)
		   (* (or letter special digit))))
      (ip       (: (+ digit) #\.
		   (+ digit) #\.
		   (+ digit) #\.
		   (+ digit)))
      )
  ((+ (or #\tab #\return #\space #\newline))
    (ignore))
  ((: #\; (* all))			; comments ('all' does not include EOL)
    (ignore))
  ((: #\? (? id))			; variables
    (let ((var-name (the-substring 1 (the-length))))
      (cons 'variable var-name)))
   ; special symbols and tokens
  (#\(    'lparen)
  (#\)    'rparen)
  (#\.    'dot)
  (#\,    'comma)
  (":-"   'conseq)
  ("says" 'says)
   ; constants
  ((: (? (in "-+")) (+ digit))
    (cons 'constant (the-fixnum)))	; leave out #o, #d, #x for now
  (id  (cons 'symb-constant (the-string)))	; symbolic constant
   ; double-quoted strings
  ((: "\"" (* (or (out #a000 #\\ #\") (: #\\ all) (: #\\ #\newline))) "\"")
    (let ((str (the-substring 1 (-fx (the-length) 1))))
      (cons 'symb-constant (escape-string str)))) ; handle the \\-escapes
  ((: "#p" ip)				; IP address
    (let* ((str (the-substring 2 (the-length)))
	   (ip-addr (build-ip-address str)))
      (if ip-addr (cons 'constant ip-addr)
	(cons 'error (list "Bad IP address: " str
		       " at position "
		       (input-port-position (the-port)))))))
  ((: "#n" ip "/" (+ digit))		; IP network
    (let* ((str (the-substring 2 (the-length)))
	    (ip-addr (build-ip-address str)))
      (if ip-addr (cons 'constant ip-addr)
	(cons 'error (list "Bad IP net address: " str
		       " at position "
		       (input-port-position (the-port)))))))
  ; EOF or a bad character
  (else
    (let ((char (the-failure)))
      (if (eof-object? char) char
	(cons 'error (list "Bad character: " char
		       " at position "
		       (input-port-position (the-port)))))))
      ))

(set! *hygien?* old-hygien)		; Restore the R5RS macro-expansion flag

; escape-string STRING -> STRING
; Take a string and return it with the follwoing substitutions:
; \\"  -> #\"
; \\n  -> #\newline
; \\\\ -> #\\
; \<newline> -> nothing
(define (escape-string str)
  (call-with-output-string
    (lambda (oport)
      (call-with-input-string str
	(lambda (iport)
	  (do ((c (read-char iport) (read-char iport)))
	      ((eof-object? c))
	    (if (char=? c #\\)
	      (let ((c (read-char iport)))
		(and (not (eof-object? c))
		  (case c
		    ((#\n) (write-char #\newline oport))
		    ((#\newline) #f)	; skip \n after #\\
		    (else (write-char c oport)))))
	      (write-char c oport))))))))


; build-ip-address STRING -> IP-address or #f
; STRING has the form "<4-quads>" or "<4-quads>/<width>"
; Return #f if parsing failed
(define (build-ip-address str)
  (call-with-input-string str
    (lambda (port)
      (and-let*
	((cnv (lambda (str) (string->integer str 0 (string-length str))))
	 (q1 (cnv (next-token '() '(#\. *eof*) "" port)))
	 ((< q1 256))
	 (_ (read-char port))
	 (q2 (cnv (next-token '() '(#\. *eof*) "" port)))
	 ((< q2 256))
	 (_ (read-char port))
	 (q3 (cnv (next-token '() '(#\. *eof*) "" port)))
	 ((< q3 256))
	 (_ (read-char port))
	 (q4 (cnv (next-token '() '(#\. #\/ *eof*) "" port)))
	 ((< q4 256))
	 (delim (read-char port))
	 (width 
	   (if (eof-object? delim) 32
	     (and (char=? delim #\/)
	       (cnv (next-token '() '(*eof*) "" port)))))
	 ((<= width 32))
	  )
	(make-ip-address
	  (if (> q1 31)			; Old Bigloo hack
	    (+ (* 256 (+ (* 256 (+ (* 256.0 q1) q2)) q3)) q4)
	    (+ (* 256 (+ (* 256 (+ (* 256 q1) q2)) q3)) q4))
	  width)))))
	

(define (test-binder-lex str expected)
  (cerr "Testing lexing of " str "..." nl)
  (let ((result
	  (call-with-input-string str
	    (lambda (port)
	      (read/rp binder-lex port)))))
    (assert (equal? result expected))))

(test-binder-lex "" (with-input-from-string "" read))
(test-binder-lex "?" '(variable . ""))
(test-binder-lex "?abc*\n" '(variable . "abc*"))
(test-binder-lex "  Abc*" '(symb-constant . "Abc*"))
(test-binder-lex "  \"Abc\\\\x\\\"\\\n\"*" '(symb-constant . "Abc\\x\""))
(test-binder-lex ";xxx\n1," '(constant . 1))
(test-binder-lex ";xxx\n. 1" 'dot) ; need a space to disambiguate from symbol
(test-binder-lex "\tsays " 'says)
(test-binder-lex "\t\"says\" " '(symb-constant . "says"))
(test-binder-lex "\"abc" '(error "Bad character: " #\" " at position " 1))
(test-binder-lex "#p10.0.0.1" `(constant . ,(make-ip-address #x0a000001 32)))
(test-binder-lex "#p10.0.257.1"
  `(error "Bad IP address: " "10.0.257.1" " at position " 12))
; (test-binder-lex "#p10.0.1.1.1"
;   `(error "Bad character: " #\# " at position " 1))
(test-binder-lex "#n10.1.0.0/24"
  `(constant . ,(make-ip-address #x0a010000 24)))
(test-binder-lex "#n10.0.0.0/36." `(error "Bad IP net address: " "10.0.0.0/36"
				   " at position " 13))


(define (test-binder-parse str expected)
  (cerr "Testing parsing of " str "..." nl)
  (let ((result
	  (call-with-input-string str
	    (lambda (port)
	      (try
		(read/lalrp binder-grammar binder-lex port)
		(lambda (esc proc msg obj)
		  (esc (make-error-assertion (list proc msg obj)))))))))
    (assert (equal? result expected))))

(test-binder-parse "" '())
(test-binder-parse "may"
  (make-error-assertion
    (list "parser" "parse error (unexpected token `*eoi*')" #eof-object)))
(test-binder-parse "may(1)"
  (make-error-assertion
    (list "parser" "parse error (unexpected token `*eoi*')" #eof-object)))
(test-binder-parse "may(\")"
  (make-error-assertion
    (list "parser" "parse error (unexpected token `error')"
      '(error "Bad character: " #\" " at position " 5))))
(test-binder-parse "may(1)."
  (list (make-assertion-clause 'may/1 #f '() '(1) '())))
(test-binder-parse "may(?,?,?)"
  (make-error-assertion
    (list "parser" "parse error (unexpected token `*eoi*')" #eof-object)))
(test-binder-parse "may(?,?,?)."
  (list (make-assertion-clause 'may/3 #f '() 
	  (list binder-anon-var binder-anon-var binder-anon-var) '())))
(test-binder-parse "may(?,_,?_)."
  (list (make-assertion-clause 'may/3 #f (list (make-binder-var "_")) 
	  (list binder-anon-var "_" (make-binder-var "_")) '())))
(test-binder-parse "may(?x,\"ok x\",?zZ)."
  (list (make-assertion-clause 'may/3 #f 
	  (list (make-binder-var "zZ") (make-binder-var "x")) 
	  (list (make-binder-var "x") "ok x" (make-binder-var "zZ")) '())))
(test-binder-parse "may(?x,?Y,?) :- application says ip(?ip)."
  (list (make-assertion-clause 'may/3 #f 
	  (list (make-binder-var "Y") (make-binder-var "x"))
	  (list (make-binder-var "x") (make-binder-var "Y") binder-anon-var)
	  (list
	    (make-assertion-pcall 'ip/1 "application" 
	      (list (make-binder-var "ip")) 
	      (list (make-binder-var "ip")))))))
(test-binder-parse "may(2,?Y,?) :- ?Y says ip(?ip,#p10.0.0.1), ok(?ip)."
  (let ((Y (make-binder-var "Y")) (ip (make-binder-var "ip")))
    (list (make-assertion-clause 'may/3 #f  `(,Y) `(2 ,Y ,binder-anon-var)
	    (list
	    (make-assertion-pcall 'ip/2  Y `(,ip ,Y)
	      (list ip (make-ip-address #x0a000001 32)))
	    (make-assertion-pcall 'ok/1 #f `(,ip) `(,ip)))))))



;------------------------------------------------------------------------
;			Post-parsing validation

; We pass the result of the parsing of an assertion through a pipeline
; of validators. Each validator either checks for some specific property, or
; performs some conversion on the result.
; If a validator finds a problem, it should return an error-assertion.
; The function binder-parse drives the pipeline.

; Check to make sure the head of any clause has the empty context
(define (test-no-context-in-head clauses)
  (or
    (and-let*
      ((bad-clause
	 (any?
	   (lambda (clause)
	     (assert (assertion-clause? clause))
	     (and (assertion-clause-ctx clause) clause))
	   clauses)))
      (make-error-assertion
	(list "test-no-context-in-head"
	  "non-empty context in the head of the clause" bad-clause)))
    clauses))


; Check to make sure each clause is safe
(define (test-clause-safety clauses)
  (or
    (and-let*
      ((bad-clause
	 (any?
	   (lambda (clause)
	     (assert (assertion-clause? clause))
	     (and-let*
	       ((fv (assertion-clause-fv clause))
		((pair? fv))
		(body (assertion-clause-body clause))
		((any?
		   (lambda (var)
		     (not
		       (any?
			 (lambda (pcall)
			   (member var (assertion-pcall-fv pcall)))
			 body)))
		   fv)))
	       clause))
	   clauses)))
      (make-error-assertion
	(list "test-clause-safety"
	  "unsafe clause" bad-clause)))
    clauses))


; A necessary (albeit not suifficient) test that the context position
; of an atom has the IN mode. That is, if the ctx of an atom is a
; variable, that variable must occur either in the head of a clause or
; in some of the previous atoms in the body.
; We also make sure that the anonymous variable never appears in the
; context.

(define (test-in-mode-of-ctx clauses)
  (or
    (and-let*
      ((error-msg
	 (any?
	   (lambda (clause)
	     (assert (assertion-clause? clause))
	     (cdr
	       (fold
	         ; seed is a pair: vars seen so far, error-msg or #f
		 (lambda (pcall seed)
		   (let ((seen-vars (car seed))
			 (error-msg (cdr seed)))
		     (if error-msg seed	; error was detected
		       (let* ((ctx (assertion-pcall-ctx pcall))
			      (error-msg
				(and (binder-var? ctx)
				  (cond
				    ((equal? binder-anon-var ctx)
				      (list "anonymous variable in context"
					clause))
				    ((not (member ctx seen-vars))
				      (list "reorder atoms in the body of"
					clause
					"so that the var in the ctx of atom"
					pcall
					"is mentioned previously"))
				    (else #f)))))
			 (cons (append (assertion-pcall-fv pcall)
				 seen-vars)
			   error-msg)))))
		 (cons (assertion-clause-fv clause) #f) ; init seed
		 (assertion-clause-body clause))))
	   clauses)))
      (make-error-assertion 
	(cons "test-in-mode-of-ctx" error-msg)))
    clauses))



; Group clauses into rule nests
; Verify that all the clauses with the same predicate are adjacent.
; To make it easy to add to the group, we reverse the clauses
(define (group-clauses clauses)
  (if (null? clauses) '()
    (let ((clauses (reverse clauses)))
      (let loop ((clauses (cdr clauses))
		 (curr-nest (list
			      (assertion-clause-name (car clauses))
			      (car clauses)))
		  (rule-nest '()))
    (if (null? clauses) (cons curr-nest rule-nest)
      (let* ((clause  (car clauses))
	     (clauses (cdr clauses))
	     (name (assertion-clause-name clause)))
	(cond
	  ((eq? name (car curr-nest)) ; the current clause continues
	    (loop clauses
	      (cons* name clause (cdr curr-nest))
	      rule-nest))
	  ((assq name rule-nest)
	    (make-error-assertion
	      (list "group-clauses"
		"clause with the same predicate is not adjacent"
		clause)))
	  (else				; the current clause starts a new nest
	    (loop clauses
	      (list name clause)
	      (cons curr-nest rule-nest))))))))))


; Test to make sure that all predictes with no context in all the bodies
; are actually defined in the current assertion. Although the reference
; to an undefined predicate is not a syntax error (it simply means
; that the corresponding clause can never be satisfied), it would help
; a user if we report such a condition early, at parsing time -- because
; the corresponding assertion is most likely meaningless.

(define (test-local-predicates-defined nests)
  (or
    (and-let*
      ((undefined-pred
	 (any?
	   (lambda (nest)
	     (any?
	       (lambda (clause)
		 (any?
		   (lambda (pcall)
		     (and-let*
		       (((not (assertion-pcall-ctx pcall)))
			(name (assertion-pcall-name pcall))
			((not (assq name nests))))
		       name))
		   (assertion-clause-body clause)))
	       (cdr nest)))
	   nests)))
      (make-error-assertion
	(list "test-local-predicates-defined"
	  "reference to an undefined local predicate" undefined-pred)))
    nests))



(define post-parsing-pipeline
  (list test-no-context-in-head
        test-clause-safety
        test-in-mode-of-ctx
        group-clauses
        test-local-predicates-defined))

; The root function of this module: binder-parse

(define (binder-parse str)
  (let ((result				; Do the parsing proper
	  (call-with-input-string str
	    (lambda (port)
	      (try
		(read/lalrp binder-grammar binder-lex port)
		(lambda (esc proc msg obj)
		  (esc (make-error-assertion (list proc msg obj)))))))))

    (let loop ((result result) (pipeline post-parsing-pipeline))
      (cond
	((error-assertion? result) result)
	((null? pipeline) result)
	(else
	  (assert (or (pair? result) (null? result)))
	  (loop ((car pipeline) result) (cdr pipeline)))))))

; Test for match
; We understand and traverse through the error assertion
; the expression to match may contain *any* wildcard

(define (test-binder-match? e1 e2)
  (let loop ((e1 e1) (e2 e2))
    (cond
      ((eq? e2 '*any*) #t)
      ((pair? e1)
	(and (pair? e2) (loop (car e1) (car e2))
	  (loop (cdr e1) (cdr e2))))
      ((error-assertion? e1)
	(and (error-assertion? e2)
	  (loop (error-assertion-msg e1) (error-assertion-msg e2))))
      (else (equal? e1 e2)))))

; The 'expected' argument may contain wildcards encoded as *any*
(define (test-binder-parse-full str expected)
  (cerr "Testing full parsing of " str "..." nl)
  (let* ((result (binder-parse str)))
    (if (not (test-binder-match? result expected))
      (assert (equal? result expected)))))

(test-binder-parse-full "" '())
(test-binder-parse-full "may(x,y,x). may(?x,?y,?) :- foo(?x), foo(?x)."
  (make-error-assertion (list "test-clause-safety" "unsafe clause" '*any*)))
(test-binder-parse-full "may(x,y,x). may(?x,?y,?) :- foo(?x), foo(?y). foo(?)."
  (list '(may/3 *any* *any*) '(foo/1 *any*)))
(test-binder-parse-full "may(x,y,x). may(?x,?y,?) :- foo(?x), foo(?y)."
  (make-error-assertion
    (list "test-local-predicates-defined"
      "reference to an undefined local predicate" 'foo/1)))
(test-binder-parse-full "may(x,y,x). may(?x,?y,?):- foo(?x),foo(?y). foo(?,1)."
  (make-error-assertion
    (list "test-local-predicates-defined"
      "reference to an undefined local predicate" 'foo/1)))
(test-binder-parse-full "may(x,y,x). foo(x). may(?x,?y,?) :- foo(?x), foo(?y)."
  (make-error-assertion (list "group-clauses"
			  "clause with the same predicate is not adjacent"
			  '*any*)))

(test-binder-parse-full 
  "foo(x). may(?x,?y,?) :- foo(?x), ? says foo(?y). may(x,y,x)."
  (make-error-assertion (list "test-in-mode-of-ctx"
			  "anonymous variable in context"
			  '*any*)))

(test-binder-parse-full 
  "foo(x). may(?x,?y,?) :- foo(?x), ?z says foo(?y). may(x,y,x)."
  (make-error-assertion (list "test-in-mode-of-ctx"
			  "reorder atoms in the body of" '*any*
			  "so that the var in the ctx of atom"
			  '*any* "is mentioned previously")))

; The assertion from the use cases
(define binder-test-init-assertion
  (apply string-append
    (list-intersperse
      (list
	"may(channel,MEMO,?) :- application says ipaddress(?IP),"
	"                       internal(?IP)."
	"may(channel,MEMO,?) :- known_user(Joe)."
	""
	"may(channel,\"DEMO-IMG\", ?Access) :-"
	"            pubkey(Dean,?Dean_key),"
	"            ?Dean_key says may(channel,\"DEMO-IMG\", ?Access)."
	""
	"internal(#p192.16.167.26)."
	"internal(?IP) :- application says ip_of(?IP,#n10.0.0.0/8)."
	""
	"known_user(?user) :- pubkey(?user,?key), pubkey_fingerprint(?key)."
	""
	"pubkey(Dean,\"abcdef\")."
	"pubkey(Joe,\"0123456789\")."
	"  ; A convenient abbreviation"
	"pubkey_fingerprint(?x) :- application says pubkey_fingerprint(?x)."
	)
      nl)))

; Test the assertion from the use cases
(test-binder-parse-full
  binder-test-init-assertion
  (list
    '(may/3 *any* *any* *any*)
    '(internal/1 *any* *any*)
    '(known_user/1 *any*)
    '(pubkey/2 *any* *any*)
    '(pubkey_fingerprint/1 *any*)))
