#! /usr/bin/env ./run-bigloo

(define docstrings
'(
"	       The Binder (Soutei) Standalone module"
""
"This is the root module of Soutei -- standalone mode. "
"It listens on a specified socket for authorization requests"
"and returns the corresponding advice. This module also implements"
"a policy for managing the Soutei cache. This module builds"
"upon binder-application."
""
"Synopsis:"
"  soutei-standalone init-config-file hostname:port"
""
"where"
"  init-config-file -- the name of the file with the initial"
"     soutei assertion."
"  hostname:port specifies the hostname and the port to listen to"
"     for requests. Normally, it should be something like"
"     localhost:1500. It is a good idea to limit potential"
"     clients to the local machine."
""
"All errors go to stderr, which ought to be re-directed into a file."
""
"Example:"
""
"  soutei-standalone binder-init-assertion.txt localhost:1500"
""
"A request should be an S-expression of the following form:"
"   (req-id (key val) ...)"
"where req-id is an identifier for the request (of no significance to soutei)"
"and (key val) pairs specify the attributes of the request,"
"e.g., (IP \"127.0.0.1\") (Subject \"user\"), etc."
""
"A response is an S-expression (req-id #t) or (req-id #f)"
"The former is an advice for an application to go ahead with"
"the request req-id."
))

; The key-value pairs from the request are passed to binder-application,
; with the following exception:
;  association (Assertion-cache "name" assertion-body)
; specifies a SOUTEI assertion to compile and cache.
;  association (Assertion-once "name" assertion-body)
; specifies an assertion to use for the current request only. 
;
; To a very slight extent, this module is tuned in for making authorization
; decisions for Metcast Channels:
;       http://www.metnet.navy.mil/Metcast/Metcast-Channels.html
; although the dependency is insignificant and can be eliminated.
; We intend to really use this module for all authorization decisions
; for Metcast server because Metcast Channels make a good demo of
; Soutei.

;
; $Id: binder-standalone.scm,v 1.2 2004/07/13 22:41:46 oleg Exp $

(module binder-standalone
  (include "myenv-bigloo.scm")
  (include "srfi-12.scm")
  ;(include "srfi-13-local.scm")
  ;(include "util.scm")
  ;(include "input-parse.scm")
  (import
    (binder-application "binder-application.scm")
  )
  (main main)
  ; (export ...)
)

; ;; Initial Setup
; (OS:putenv "TZ" "GMT0")	  ;; All time is GMT0, not dependent on locale.


		; Takes the request, logs it. It returns the
		; request in the s-expression format
		; (req-id (key value) ...)
(define (take-request in-port)

  (define (log-request req)
    (cerr "\n----> New "
      " request" nl req nl))

  (let ((req (read in-port)))
    (log-request req)
    (assert (pair? req))
    req)
)

(define (main-loop ns-cache socket)
  (socket-accept-connection socket) ; wait for a connection
  (handle-exceptions
    exn
    (let ((err-msgs
	    (if ((condition-predicate 'exn) exn)
	      (list "Runtime error:"
		((condition-property-accessor 'exn 'message) exn))
	      (list "Other exception:" exn))))
      (cerr "******> Runtime error or exception" err-msgs)
      (socket-shutdown socket #f)
      (main-loop ns-cache socket))

    (let*-values
      (((request) (take-request (socket-input socket)))
       ((req-id req-alist) (values (car request) (cdr request)))

	; compile the permanent assertions, Assertion-cache,
	; and extend the cache
       ((ns-cache req-alist)
	 (let loop ((ns-cache ns-cache) (curr req-alist) (new '()))
	   (cond
	     ((null? curr) (values ns-cache new))
	     ((eq? 'Assertion-cache (caar curr))
	       (let*-values
		 (((key ns-name ns-body) (apply values (car curr))))
		 (loop
		   (or
		     (binder-application-extend ns-cache ns-name ns-body)
		     ns-cache)
		   (cdr curr)
		   new)))
	     (else (loop ns-cache (cdr curr) (cons (car curr) new))))))

	; compile once assertions, Assertion-once,
	; and extend the cache
       ((ns-cache-local req-alist)
	 (let loop ((ns-cache-local ns-cache) (curr req-alist) (new '()))
	   (cond
	     ((null? curr) (values ns-cache-local new))
	     ((eq? 'Assertion-once (caar curr))
	       (let*-values
		 (((key ns-name ns-body) (apply values (car curr))))
		 (loop
		   (or
		     (binder-application-extend ns-cache-local ns-name ns-body)
		     ns-cache-local)
		   (cdr curr)
		   new)))
	     (else (loop ns-cache-local (cdr curr) (cons (car curr) new))))))

		   
	; Run the query
	((_) (cerr nl "Running the query" nl))
	((result) (binder-application-query ns-cache-local req-alist))
	)

      (cerr nl "Result: " result nl)
      (write (list req-id result) (socket-output socket))
      (display "\n" (socket-output socket))
      (flush-output-port (socket-output socket))
      (socket-shutdown socket #f)
      (main-loop ns-cache socket)
      )))


;------------------------------------------------------------------------
; 				Root module

(define (main argv)

  (define (help msg)
    (cerr nl msg nl)
    (for-each (lambda (str) (cerr str nl)) docstrings)
    (exit 4))


  (cerr "argv: " argv nl)

  (let*
      ((argv (cdr argv))	; (car argv) is program's name, as usual
       (_ (or (>= (length argv) 2)
	      (help "At least two arguments are needed")))
	; the current version of bigloo retains the interpreter options
	; as well. A bummer!
	; We pick the last two arguments
       (argv-len (length argv))
       (init-assertion-file (list-ref argv (- argv-len 2)))
       (listen-addr (list-ref argv (- argv-len 1)))
       (_ (cerr nl "******> SOUTEI " init-assertion-file " " listen-addr nl))

       (parsed-addr (string-split listen-addr '(#\:)))
       (host-addr (and (pair? parsed-addr) (car parsed-addr)))
       (listen-port
	 (and (pair? parsed-addr) (pair? (cdr parsed-addr))
	   (null? (cddr parsed-addr))
	   (string->integer (cadr parsed-addr) 0
	     (string-length (cadr parsed-addr)))))
       (_ (or listen-port 
	    (help "The second argument must be hostname:port")))
       (init-ns-cache 
	 (binder-application-init init-assertion-file))
       (socket (make-server-socket listen-port))
       )

    (cerr nl "******> SOUTEI Ready" nl)
    (main-loop init-ns-cache socket)
    (cerr nl "******> SOUTEI Finished" nl)
))

